from typing import Dict, List

from loguru import logger

from dibber.images import docker_tag, get_config


def validate(images: Dict[str, List[str]]):
    for image in images.keys():
        tags = {}
        for version in images[image]:
            name = f"{image}/{version}"
            logger.debug("Validating {name}", name=name)
            _validate_version(image, version)
            _tags = _get_tags(image, version)

            # Check tags are unique
            for tag in _tags:
                if tag in tags.keys():
                    src = tags[tag]
                    raise ValueError(
                        f"{name} trying to redefine {tag} already defined in {src}"
                    )
                tags[tag] = name
                logger.debug("{name} has tag {tag}", name=name, tag=tag)


def _validate_version(image: str, version: str):
    pass


def _get_tags(image: str, version: str):
    tags = [docker_tag(image, version)]
    config = get_config(image, version)
    tags += [docker_tag(image, tag) for tag in config.tags]
    return tags



import multiprocessing.context
import os
import signal
import sys
from enum import Enum
from multiprocessing import Pool
from typing import Optional

import click

import dibber.utils as utils
from dibber.images import (
    build_image,
    build_image_multiplatform,
    docker_tag,
    find_images,
    scan_image,
    sort_images,
    update_scanner,
    upload_tags_from_local_registry,
)
from dibber.settings import conf
from dibber.validation import validate


class Platform(str, Enum):
    LINUX_AMD64 = "linux/amd64"
    LINUX_ARM64 = "linux/arm64"


ALL_PLATFORMS = [Platform.LINUX_AMD64.value, Platform.LINUX_ARM64.value]


def init_pool(logger_, env):
    utils.logger = logger_
    os.environ.update(env)


def _build_images(pool, images, multiplatform, platform):
    if multiplatform:
        res = pool.starmap_async(
            build_image_multiplatform,
            [(image, version, ALL_PLATFORMS, False) for image, version in images],
        )
    else:
        res = pool.starmap_async(
            build_image,
            [(image, version, False, platform) for image, version in images],
        )

    while True:
        try:
            # Have a timeout to be non-blocking for signals
            res.get(0.25)
            break
        except multiprocessing.context.TimeoutError:
            pass


def _build_all_images(
    parallel: int, multiplatform: bool, platform: Optional[Platform] = None
):
    platform = platform.value if platform else None
    images = find_images()
    validate(images)
    sorted_images = sort_images(images)

    if parallel == 1:
        images = [img_conf.image for img_conf in sorted_images]
        for image, version in images:
            if multiplatform:
                build_image_multiplatform(image, version, ALL_PLATFORMS)
            else:
                build_image(image, version, platform=platform)
    else:
        utils.logger.info(f"Building {len(sorted_images)} images in {parallel} threads")
        utils.logger.remove()
        utils.logger.add(sys.stderr, enqueue=True, level="INFO")

        original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
        with Pool(
            parallel, initializer=init_pool, initargs=(utils.logger, os.environ)
        ) as pool:
            signal.signal(signal.SIGINT, original_sigint_handler)

            max_prio = max(ic.priority for ic in sorted_images)
            for prio in range(1, max_prio + 1):
                images = [
                    img_conf.image
                    for img_conf in sorted_images
                    if img_conf.priority == prio
                ]
                try:
                    utils.logger.info(
                        "Building {c} priority {prio} images with up to {parallel} threads",
                        c=len(images),
                        prio=prio,
                        parallel=parallel,
                    )
                    _build_images(pool, images, multiplatform, platform)
                except KeyboardInterrupt:
                    utils.logger.error("Caught KeyboardInterrupt, terminating workers")
                    pool.terminate()
                    raise

            pool.close()


@click.group(help="Manage docker images")
def cli():
    pass


@cli.command(help="Build docker images")
@click.option(
    "--parallel",
    default=2,
    type=int,
    help="Number of parallel images to build.",
    show_default=True,
)
@click.option(
    "--platform",
    default=None,
    type=Platform,
    help="Platform to build for.",
)
def build(parallel: int, platform: Optional[Platform]):
    _build_all_images(parallel, platform=platform, multiplatform=False)


@cli.command(help="Build docker images for multiple platforms")
@click.option(
    "--parallel",
    default=2,
    type=int,
    help="Number of parallel images to build.",
    show_default=True,
)
def build_multiplatform(
    parallel: int,
):
    _build_all_images(parallel, multiplatform=True)


@cli.command(help="Upload docker tags")
def upload():
    images = find_images()
    validate(images)
    upload_tags_from_local_registry(images)


@cli.command(help="Scan docker images")
def scan():
    update_scanner()
    images = find_images()
    vuln_images = []
    for image, versions in sorted(images.items()):
        for version in versions:
            if not scan_image(image, version):
                vuln_images.append(docker_tag(image, version))

    if vuln_images:
        utils.logger.error("Some images have vulnerabilities!")
        for img in vuln_images:
            print(f" - {img}")

        raise sys.exit(1)


@cli.command(help="List unique docker images managed by this tool")
def list():
    images = find_images()
    for image, versions in images.items():
        for version in versions:
            print(docker_tag(image, version))


@cli.command(help="Get the configured Docker username")
def docker_username():
    print(conf.docker_user)



import time
from datetime import timedelta
from pathlib import Path
from typing import Dict, List, Optional

import humanize
from loguru import logger
from pydantic import BaseModel
from yaml import load

from dibber.settings import conf
from dibber.utils import run

try:
    from yaml import CLoader as Loader
except ImportError:
    from yaml import Loader


class Config(BaseModel):
    tags: List[str]


def find_images() -> Dict[str, List[str]]:
    result = {}

    images = [
        p.name for p in Path(".").iterdir() if p.is_dir() and not p.name.startswith(".")
    ]

    for image in images:
        versions = [
            p.name
            for p in Path(image).iterdir()
            if p.is_dir() and not p.name.startswith(".")
        ]
        result[image] = versions

    return result


class ImageConf:
    priority: int
    image: list[str]

    def __init__(self, priority: int, image: list[str]):
        self.priority = priority
        self.image = image

    def __repr__(self):
        return f"<{':'.join(self.image)} @ {self.priority} prio>"


def sort_images(images_: Dict[str, List[str]]) -> List[ImageConf]:
    images = []
    for image, versions in images_.items():
        for version in versions:
            images.append(f"{image}/{version}")
    images.sort()
    for image_or_list in conf.priority_builds:
        if isinstance(image_or_list, str):
            try:
                images.remove(image_or_list)
            except ValueError:
                logger.error(
                    "{image} found in PRIORITY_BUILDS is incorrect", image=image_or_list
                )
                raise
        else:
            for _img in image_or_list:
                try:
                    images.remove(_img)
                except ValueError:
                    logger.error(
                        "{image} found in PRIORITY_BUILDS is incorrect", image=_img
                    )
                    raise

    priority = 1
    result = []
    for image_or_list in conf.priority_builds:
        if isinstance(image_or_list, str):
            try:
                result.append(
                    ImageConf(
                        priority=priority, image=image_or_list.split("/", maxsplit=1)
                    )
                )
            except ValueError:
                logger.error(
                    "{image} found in PRIORITY_BUILDS is incorrect", image=image_or_list
                )
                raise
        else:
            for _img in image_or_list:
                try:
                    result.append(
                        ImageConf(priority=priority, image=_img.split("/", maxsplit=1))
                    )
                except ValueError:
                    logger.error(
                        "{image} found in PRIORITY_BUILDS is incorrect", image=_img
                    )
                    raise
        priority += 1

    result += [
        ImageConf(priority=priority, image=img.split("/", maxsplit=1)) for img in images
    ]

    return result


def build_image(image: str, version: str, verbose=True, platform: Optional[str] = None):
    config = get_config(image, version)
    name = f"{image}/{version}"
    tag = docker_tag(image, version)

    logger.info("Building {name}", name=name)

    # Full commandline with all tags in one
    cmd = ["docker", "build", name, "-t", tag]
    for tag in config.tags:
        full_name = docker_tag(image, tag)
        cmd += ["-t", full_name]

    # make it possible to reuse this image in other local builds
    cmd += ["-t", docker_tag(image, tag=version, local=True)]

    if platform:
        cmd += ["--platform", platform]

    start = time.perf_counter()
    run(cmd, verbose=verbose)
    end = time.perf_counter()
    if not verbose:
        logger.info(
            "Built {name} in {elapsed}",
            name=name,
            elapsed=humanize.precisedelta(timedelta(seconds=end - start)),
        )


def build_image_multiplatform(
    image: str, version: str, platforms: List[str], verbose=True
):
    name = f"{image}/{version}"

    logger.info("Building {name}", name=name)

    localhost_tag = f"{conf.local_registry}/{image}:{version}"
    build_cmd = [
        "docker",
        "buildx",
        "build",
        "--platform",
        ",".join(platforms),
        name,
        "-t",
        localhost_tag,
        "--build-arg",
        f"LOCAL_REGISTRY={conf.local_registry}/",
        "--network=host",
        "--push",  # push to localhost registry
    ]

    start = time.perf_counter()
    run(build_cmd, verbose=verbose)
    end = time.perf_counter()
    if not verbose:
        logger.info(
            "Built {name} in {elapsed}",
            name=name,
            elapsed=humanize.precisedelta(timedelta(seconds=end - start)),
        )


def upload_tags_from_local_registry(images: Dict[str, List[str]]):
    # local docker registry runs by HTTP, so we state it in regctl
    run(["regctl", "registry", "set", "--tls", "disabled", conf.local_registry])
    for image, versions in images.items():
        for version in versions:
            config = get_config(image, version)
            localhost_tag = f"{conf.local_registry}/{image}:{version}"
            tags = [docker_tag(image, tag) for tag in config.tags]
            tags.append(docker_tag(image, version))
            for tag in tags:
                run(["regctl", "image", "copy", localhost_tag, tag], verbose=True)


def docker_image(image: str) -> str:
    return f"{conf.docker_user}/{image}"


def docker_tag(image: str, tag: str, local: bool = False) -> str:
    if not local:
        return f"{docker_image(image)}:{tag}"
    return f"{image}:{tag}"


def get_config(image: str, version: str) -> Config:
    config_path = f"{image}/{version}/config.yaml"
    config_text = Path(config_path).read_text(encoding="utf-8")
    config = load(config_text, Loader=Loader)
    return Config(**config)


def update_scanner():
    logger.info("Updating trivy database")
    run(["trivy", "image", "--download-db-only"])


def scan_image(image: str, version: str) -> bool:
    try:
        run(
            [
                "trivy",
                "image",
                "--skip-update",
                "--severity",
                "HIGH,CRITICAL",
                "--exit-code",
                "1",
                "--timeout",
                "7m",
                f"{docker_image(image)}:{version}",
            ],
            cwd=f"{image}/{version}",
        )
        return True
    except Exception:
        logger.error(
            "{image}:{version} has vulnerabilities!", image=image, version=version
        )
        return False



from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    TomlConfigSettingsSource,
)

PriorityBuilds = list[list[str]]


# The properties in dibber.toml should have matching names
class Settings(BaseSettings):
    # User or organization name, used for "user/image" -name generation, optionally leading registry hostname
    docker_user: str

    local_registry: str = ""

    # List of images that should be built beforehand
    priority_builds: PriorityBuilds = []

    model_config = SettingsConfigDict(toml_file="dibber.toml")

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (TomlConfigSettingsSource(settings_cls),)


conf = Settings()



from subprocess import PIPE, STDOUT, check_call  # nosec
from subprocess import run as sp_run  # nosec
from typing import List

from loguru import logger

_ = logger


def run(cmd: List[str], verbose=True, cwd=None):
    logger.debug("Executing: {cmd}", cmd=" ".join(cmd))
    if verbose:
        check_call(cmd, cwd=cwd)  # nosec B603
    else:
        result = sp_run(cmd, stdout=PIPE, stderr=STDOUT, cwd=cwd)  # nosec B603
        if result.returncode > 0:
            logger.error(
                "{cmd} exited with code {code}",
                cmd=" ".join(cmd),
                code=result.returncode,
            )
            logger.error(result.stdout.decode("utf-8"))
            raise Exception(f"Failed to run {cmd}")
